<?php

namespace Drupal\denormalizer\Entity;

use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Url;
use Drupal\webform\WebformSubmissionInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Provides an interface defining a denormalizer_table information entity.
 *
 * @ConfigEntityType(
 *   id = "denormalizer_table",
 *   label = @Translation("Denormalizer table"),
 *   label_collection = @Translation("Denormalizer tables"),
 *   label_singular = @Translation("denormalizer table"),
 *   label_plural = @Translation("denormalizer tables"),
 *   label_count = @PluralTranslation(
 *     singular = "@count denormalizer table",
 *     plural = "@count denormalizer tables",
 *   ),
 *   handlers = {
 *     "list_builder" = "Drupal\denormalizer\DenormalizerTableListBuilder",
 *     "form" = {
 *       "default" = "Drupal\denormalizer\Form\DenormalizerTableForm",
 *       "add" = "Drupal\denormalizer\Form\DenormalizerTableForm",
 *       "edit" = "Drupal\denormalizer\Form\DenormalizerTableForm",
 *       "delete" = "Drupal\Core\Entity\EntityDeleteForm",
 *     },
 *     "route_provider" = {
 *       "html" = "Drupal\Core\Entity\Routing\AdminHtmlRouteProvider"
 *     },
 *   },
 *   admin_permission = "administer denormalizer",
 *   config_prefix = "denormalizer_table",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label",
 *     "uuid" = "uuid",
 *     "status" = "status",
 *     "source" = "source",
 *   },
 *   config_export = {
 *     "id",
 *     "label",
 *     "status",
 *     "source",
 *     "configuration",
 *   },
 *   links = {
 *     "add-form" = "/admin/structure/denormalizer-tables/add",
 *     "edit-form" = "/admin/structure/denormalizer-tables/manage/{denormalizer_table}",
 *     "delete-form" = "/admin/structure/denormalizer-tables/manage/{denormalizer_table}/delete",
 *     "collection" =  "/admin/structure/denormalizer-tables"
 *   }
 * )
 */
class DenormalizerTable extends ConfigEntityBase implements DenormalizerTableInterface {

  /**
   * The denormalizer table ID.
   *
   * @var string
   */
  protected $id;

  /**
   * The denormalizer table label.
   *
   * @var string
   */
  protected $label;

  /**
   * The source ID.
   *
   * @var string
   */
  protected $source;

  /**
   * The source configuration.
   *
   * @var array
   */
  protected $configuration = [];

  protected $entityFieldManager;

  protected $entityTypeManager;

  protected $entityFieldTypeManager;

  protected $databaseConnection;

  protected $entityFieldDefinitions = [];

  /**
   * {@inheritdoc}
   */
  public function getSourceId() {
    return $this->source;
  }

  /**
   * {@inheritdoc}
   */
  public function setSourceId($source_id) {
    // @todo validate source id, currently only supporting "entity" and
    // "non_entity".
    $this->source = $source_id;
    $this->configuration = [];
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getSourceConfiguration() {
    if ($this->getSourceId() === 'entity' && isset($this->configuration['entity_type'])) {
      $this->configuration['schema_module'] = $this->entityTypeManager()
        ->getDefinition($this->configuration['entity_type'])
        ->getProvider();
    }
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setSourceConfiguration(array $configuration) {
    if (isset($configuration['fields']['base'])) {
      $configuration['fields']['base'] = array_filter($configuration['fields']['base'], function ($value) {
        return !empty($value);
      });
    }
    if (isset($configuration['fields']['bundle'])) {
      $configuration['fields']['bundle'] = array_filter($configuration['fields']['bundle'], function ($value) {
        return !empty($value);
      });
    }

    if (!empty($configuration['entity_type']) && empty($configuration['fields']['bundle']) && !empty($configuration['bundle'])) {
      if ($configuration['entity_type'] !== 'webform_submission') {
        $bundle_fields = array_keys(array_filter($this->entityFieldDefinitions(), function (FieldDefinitionInterface $field_definition) {
          return !empty($field_definition->getTargetBundle());
        }));
        $configuration['fields']['bundle'] = array_combine($bundle_fields, $bundle_fields);
      }
      else {
        /** @var \Drupal\webform\WebformEntityStorageInterface $webform_storage */
        $webform_storage = $this->entityTypeManager()
          ->getStorage('webform');
        $webform = $webform_storage->load($configuration['bundle']);
        $webform_elements = array_keys($webform->getElementsInitializedFlattenedAndHasValue());
        $configuration['fields']['bundle'] = array_combine($webform_elements, $webform_elements);
      }
    }
    $this->configuration = $configuration;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDbTableName() {
    return 'denormalizer_' . $this->id();
  }

  /**
   * {@inheritdoc}
   */
  public function getBaseTable() {
    $configuration = $this->getSourceConfiguration();
    if ($this->getSourceId() === 'entity') {
      return $this->entityTypeManager()
        ->getDefinition($configuration['entity_type'])
        ->getBaseTable();
    }
    return $configuration['base_table'] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function delete() {
    parent::delete();
    $this->databaseConnection()->schema()
      ->dropTable($this->getDbTableName());
  }

  /**
   * {@inheritdoc}
   */
  public function set($property_name, $value) {
    // Invoke the setters to clear related properties.
    if ($property_name == 'source') {
      return $this->setSourceId($value);
    }
    elseif ($property_name == 'configuration') {
      return $this->setSourceConfiguration($value);
    }
    else {
      return parent::set($property_name, $value);
    }
  }

  public static function initBatch(DenormalizerTableInterface $dn_table, $entity_type_id, &$context) {
    $context['results']['entity'] = $dn_table;
    $context['results']['entity_type_id'] = $entity_type_id;
    $context['results']['processed'] = 1;
    $context['results']['denormalizer_table_initialized'] = $dn_table->createDatabaseTable(TRUE);
  }

  public static function populateTableBatch($entity_ids, &$context) {
    $dn_table = $context['results']['entity'] ?? NULL;
    $entity_type_id = $context['results']['entity_type_id'] ?? '';
    $entity_type_manager = \Drupal::entityTypeManager();
    $entity_storage = $entity_type_id ? $entity_type_manager->getStorage($entity_type_id) : NULL;
    if (!($dn_table instanceof DenormalizerTableInterface) || !($entity_storage instanceof EntityStorageInterface)) {
      $context['results']['processed']++;
      return;
    }

    foreach ($entity_ids as $entity_id) {
      $entity = $entity_storage->load($entity_id);
      $dn_table->insertDataFromEntityIntoDbTable($entity);
    }
    $context['results']['processed']++;
  }

  public static function finishBatch($success, $results) {
    $entity = $results['entity'];
    \Drupal::messenger()->addMessage(t('Saved the %label denormalizer table.', [
      '%label' => $entity->label(),
    ]));
    $url = new Url('entity.denormalizer_table.collection');
    return new RedirectResponse($url->toString());
  }

  public function createDatabaseTable(bool $init = FALSE) {
    $table_name = $this->getDbTableName();
    // Always prefix denormalized created tables so that they can be identified.
    $schema = $this->databaseConnection()->schema();
    $table_exists = $schema->tableExists($table_name);
    // Don't initialize anything if table exists already and this is not an
    // initialization requests.
    if ($table_exists && !$init) {
      return FALSE;
    }
    elseif ($table_exists) {
      $this->databaseConnection()->schema()
        ->dropTable($this->getDbTableName());
    }

    $schema->createTable($table_name, [
      'description' => $this->label(),
      'fields' => $this->getDbTableFieldsSchema(),
      'indexes' => $this->getDbTableIndexes(),
    ]);

    return $schema->tableExists($table_name);
  }

  public function insertDataFromEntityIntoDbTable(EntityInterface $entity) {
    try {
      $return_value = $this->databaseConnection()
        ->insert($this->getDbTableName())
        ->fields($this->getEntryFromEntity($entity))
        ->execute();
    }
    catch (\Exception $e) {
      // @todo either send to Drupal log or use a route to only apply on Denormalizer table page
      //  \Drupal::messenger()->addMessage(t('Insert failed. Message = %message', [
      //    '%message' => $e->getMessage(),
      //  ]), 'error');
    }

    return $return_value ?? NULL;
  }

  public function updateDataFromEntityIntoDbTable(EntityInterface $entity) {
    $entry = $this->getEntryFromEntity($entity);
    try {
      $entity_id_key = $entity->getEntityType()->getKey('id');
      $count = $this->databaseConnection()
        ->update($this->getDbTableName())
        ->fields($entry)
        ->condition($this->getDbTableColumnNameFromEntityField($entity_id_key), $entity->id())
        ->execute();
    }
    catch (\Exception $e) {
      // @todo either send to Drupal log or use a route to only apply on Denormalizer table page
      //      \Drupal::messenger()->addMessage($this->t('Update failed. Message = %message, query= %query', [
      //          '%message' => $e->getMessage(),
      //          '%query' => $e->query_string,
      //        ]
      //      ), 'error');
    }
    return $count ?? 0;
  }

  public function deleteDataFromEntityIntoDbTable(EntityInterface $entity) {
    $entity_id_key = $entity->getEntityType()->getKey('id');
    $this->databaseConnection()
      ->delete($this->getDbTableName())
      ->condition($this->getDbTableColumnNameFromEntityField($entity_id_key), $entity->id())
      ->execute();
  }

  public function getDbTableFieldsSchema() {
    $fields_schema = [];
    // Let support entity only for now.
    if ($this->getSourceId() !== 'entity') {
      return $fields_schema;
    }
    $is_webform_submission = $this->isWebformSubmission();
    $fields_schema = $this->getDbTableColumnsSchemaForEntityFields();
    if ($is_webform_submission) {
      $fields_schema += $this->getDbTableColumnsSchemaForWebformElements();
    }
    return $fields_schema;
  }

  public function isWebformSubmission() {
    return $this->configuration['entity_type'] === 'webform_submission';
  }

  public function getEntitySelectedFields(bool $base_field_only = FALSE) {
    $selected_fields = $this->configuration['fields']['base'];
    if (!$this->isWebformSubmission() && !$base_field_only) {
      $selected_fields += $this->configuration['fields']['bundle'];
    }
    return $selected_fields;
  }

  public function getWebformSubmissionSelectedElements() {
    /** @var \Drupal\webform\WebformEntityStorageInterface $webform_storage */
    $webform_storage = $this->entityTypeManager()
      ->getStorage('webform');
    $webform = $webform_storage->load($this->configuration['bundle']);
    return $webform->getElementsInitializedFlattenedAndHasValue();
  }

  public function countRows() {
    if (!$this->databaseConnection()->schema()->tableExists($this->getDbTableName())) {
      return 0;
    }
    $query =$this->databaseConnection()
      ->select($this->getDbTableName(), 't');
    $query->addExpression('COUNT(*)');
    return $query->execute()->fetchField() ?? 0;
  }

  public static function onContentEntityCRUD(EntityInterface $entity, bool $is_new = FALSE, bool $delete = FALSE) {
    $dn_table_storage = \Drupal::entityTypeManager()
      ->getStorage('denormalizer_table');
    $ids = $dn_table_storage->getQuery()
      ->condition('source', 'entity')
      ->execute();

    if ($ids) {
      foreach ($ids as $id) {
        /** @var \Drupal\denormalizer\Entity\DenormalizerTableInterface $dn_table */
        $dn_table = $dn_table_storage->load($id);
        $configuration = $dn_table->getSourceConfiguration();
        if ($configuration['entity_type'] === $entity->getEntityTypeId() && $configuration['bundle'] === $entity->bundle()) {
          if ($is_new) {
            $dn_table->insertDataFromEntityIntoDbTable($entity);
          }
          elseif ($delete) {
            $dn_table->deleteDataFromEntityIntoDbTable($entity);
          }
          else {
            $dn_table->updateDataFromEntityIntoDbTable($entity);
          }
        }
      }
    }
  }

  private function extractDataFromEntityForSelectedFields(EntityInterface $entity) {
    $data = [];
    foreach ($this->getEntitySelectedFields(FALSE) as $field) {
      $field_item = $entity->get($field);
      if (!$field_item->isEmpty()) {
        $field_values = $field_item->getValue();
        $property = key($field_values[0]);
        if (is_countable($field_values) && count($field_values) > 1) {
          // If multiple values separate by a comma.
          $data[$field . '__' . $property] = implode(',', array_column($field_values, $property));
        }
        else {
          $data[$field . '__' . $property] = $field_values[0][$property];
        }
      }
    }
    return $data;
  }

  private function getWebformElementOptions(array $element) {
    if ($element['#type'] === 'webform_likert') {

      return $element['#questions'];
    }

    return $element['#options'];
  }

  private function extractDataFromWebformSubmissionForSelectedElements(WebformSubmissionInterface $submission) {
    $webform_elements = $this->getWebformSubmissionSelectedElements();
    $data = [];
    foreach ($this->configuration['fields']['bundle'] as $element_key) {
      $element_data = $submission->getElementData($element_key);
      if ($element_data !== NULL) {
        $element = $webform_elements[$element_key];
        if ($this->webformElementIsMultiple($element) && is_array($element_data)) {
          $is_likert = $element['#type'] === 'webform_likert';
          foreach ($this->getWebformElementOptions($element) as $option_key => $option_value) {
            if (!$is_likert && in_array($option_key, $element_data)) {
              $data[$element_key . '__' . $option_key] = $option_key;
            }
            elseif ($is_likert && isset($element_data[$option_key])) {
              $data[$element_key . '__' . $option_key] = $element_data[$option_key];
            }
          }
        }
        else {
          $data[$element_key] = $submission->getElementData($element_key);
        }
      }
    }

    return $data;
  }

  /**
   * Gets table columns schema.
   *
   * @return array
   *   The columns schemas.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getDbTableColumnsSchemaForEntityFields() {
    $field_definitions = $this->entityFieldDefinitions();
    $selected_fields = $this->getEntitySelectedFields(FALSE);

    $columns_schema = [];
    foreach ($field_definitions as $field_definition) {
      $field_name = $field_definition->getName();
      if (!in_array($field_name, $selected_fields)) {
        continue;
      }

      $field_type_definition = $this->entityFieldTypeManager()
        ->getDefinition($field_definition->getType());
      /** @var \Drupal\Core\Field\FieldItemInterface $field_type_class */
      $field_type_class = $field_type_definition['class'];
      $schemas = $field_type_class::schema($field_definition->getFieldStorageDefinition());
      if (isset($schemas['columns'])) {
        foreach ($schemas['columns'] as $column => $schema) {
          $schema['description'] = $column . ' for ' . $field_definition->getLabel() . ' field';
          if ($column === 'target_id') {
            $schema['type'] = 'varchar_ascii';
            $schema['length'] = 255;
            unset($schema['unsigned']);
          }
          $columns_schema[$field_name . '__' . $column] = $schema;
        }
      }
    }
    return $columns_schema;
  }

  private function getDbTableColumnsSchemaForWebformElements() {
    $webform_elements = $this->getWebformSubmissionSelectedElements();
    $columns_schema = [];
    foreach ($this->configuration['fields']['bundle'] as $element_key) {
      $element = $webform_elements[$element_key];
      if ($this->webformElementIsMultiple($webform_elements[$element_key])) {
        foreach ($this->getWebformElementOptions($element) as $option_key => $option_value) {
          $description = (string) $element['#title'] . ' ' . $option_value . ' field';
          $columns_schema[$element_key . '__' . $option_key] = $this->webformElementDbTableSchema($element['#type'], $description);
        }
      }
      else {
        // All webform elements are getting a medium text field type for now.
        // @todo find out a way to give them a proper schema based on field type.
        $description = (string) $element['#title'] . ' field';
        $columns_schema[$element_key] = $this->webformElementDbTableSchema($element['#type'], $description);
      }
    }
    return $columns_schema;
  }

  private function getDbTableIndexes() {
    $table_indexes = [];
    if (!empty($this->configuration['fields']['base'])) {
      foreach ($this->configuration['fields']['base'] as $base_field) {
        if ($column_field = $this->getDbTableColumnNameFromEntityField($base_field)) {
          $table_indexes[$column_field] = [$column_field];
        }
      }
    }
    return $table_indexes;
  }

  /**
   * Helper method to instantiate the entity field manager.
   *
   * @return \Drupal\Core\Entity\EntityFieldManagerInterface
   *   The entity field manager.
   */
  private function entityFieldManager() {
    if (empty($this->entityFieldManager)) {
      $this->entityFieldManager = \Drupal::service('entity_field.manager');
    }
    return $this->entityFieldManager;
  }

  /**
   * Helper method to instantiate the database connection.
   *
   * @return \Drupal\Core\Database\Connection
   *   The database connection.
   */
  private function databaseConnection() {
    if (empty($this->databaseConnection)) {
      $this->databaseConnection = \Drupal::service('database');
    }
    return $this->databaseConnection;
  }

  private function entityFieldDefinitions() {
    if (empty($this->entityFieldDefinitions) && ($this->getSourceId() === 'entity' && !empty($this->configuration['entity_type']) && !empty($this->configuration['bundle']))) {
      $this->entityFieldDefinitions = $this->entityFieldManager()
        ->getFieldDefinitions($this->configuration['entity_type'], $this->configuration['bundle']);
    }
    return $this->entityFieldDefinitions;
  }

  /**
   * Helper method to initialize the filed type plugin manager.
   *
   * @return \Drupal\Core\Field\FieldTypePluginManagerInterface
   *   The field type manager.
   */
  private function entityFieldTypeManager() {
    if (empty($this->entityFieldTypeManager)) {
      $this->entityFieldTypeManager = \Drupal::service('plugin.manager.field.field_type');
    }
    return $this->entityFieldTypeManager;
  }

  private function webformElementDbTableSchema(string $type, $description) {
    return [
      'type' => 'text',
      'size' => $type === 'webform_computed_twig' ? 'big' : 'medium',
      'description' => $description,
    ];
  }

  private function webformElementIsMultiple(array $element) {
    $element_type = $element['#type'];
    $is_multiple = ($element_type === 'checkboxes' || !empty($element['#multiple'])) && !empty($element['#options']);
    if (!$is_multiple && $element_type === 'webform_likert' && !empty($element['#questions'])) {
      $is_multiple = TRUE;
    }

    return $is_multiple;
  }

  private function getEntryFromEntity(EntityInterface $entity) {
    $entry = $this->extractDataFromEntityForSelectedFields($entity);
    if ($this->isWebformSubmission()) {
      $entry += $this->extractDataFromWebformSubmissionForSelectedElements($entity);
    }
    return $entry;
  }

  private function getDbTableColumnNameFromEntityField(string $field_name) {
    $field_definitions = $this->entityFieldDefinitions();
    $field_definition = $field_definitions[$field_name];
    $field_type_definition = $this->entityFieldTypeManager()
      ->getDefinition($field_definition->getType());
    $field_type_class = $field_type_definition['class'];
    $column_name = '';
    if ($column = $field_type_class::mainPropertyName()) {
      $column_name = $field_name . '__' . $column;
    }
    return $column_name;
  }

}
